import PubSub from "pubsub-js";
import { EventActions } from "../../game/constants";
import { MatchSession } from "../../game/MatchSession";
import { SessionManager } from "../managers/SessionManager";
import { PlayerNotificationService } from "./PlayerNotificationService";
import { ServiceBase } from "./ServiceBase";
import { PlayerMove } from "../../game/entities/PlayerMove";

export class InteractionService extends ServiceBase{
  private sessionManager: SessionManager = new SessionManager();  
  private notifyService = new PlayerNotificationService();  

  async handleClientEventWithAck(data: any): Promise<any> {
    const { event, data: eventData } = data;
    this.logger.trace(`Handling event: ${event} with ack`);
    switch(event) {
      case EventActions.START_SESSION:
        return this.onStartSession(eventData);
      default:
        // PubSub.publish(event, eventData);
        break;
    }
  }

  handleClientEvent(data: any): any {
    const { event, data: eventData } = data;
    this.logger.trace(`Handling event: ${event}`);
    switch(event) {
      case 'client:player-move':
        this.onClientMoveResponse(eventData);
        break;
      case 'client:set-client-ready':
        this.onClientReady(eventData);
        break;
      case EventActions.PLAYER_READY:
        this.onPlayerReady(eventData);
        break;
      default:
        PubSub.publish(event, eventData);
        break;
    } 
  }

  private onStartSession(data: any): any {
    const sessionId: string = data.sessionId;
    const session: MatchSession | undefined = this.sessionManager.getSession(sessionId);

    if (session === undefined) {
      return ({
        status: 'error',
        message: 'Session not found'
      });
    } else if (session.matchInProgress) {
      return {
        status: 'error',
        message: 'Game already in progress'
      };
    } else {
      const missingHumans = session.maxPlayers - session.numPlayers;
      for (let i = 0; i < missingHumans; i++) {
        session.addPlayerToSession(session.createPlayerAI(i));
      }
      this.notifyService.sendEventToPlayers('server:match-starting', session.players);
      session.start();      
      return {
        status: 'ok'
      };
    }
  }

  public playerMove(data: any) {
    this.onClientMoveResponse(data);
  }

  private onClientMoveResponse(data: any): any {
    const { sessionId, move }: { sessionId: string, move: PlayerMove } = data;
    const session: MatchSession | undefined = this.sessionManager.getSession(sessionId);
    if (session !== undefined) {
      session.playerMove(move);
      return {
        status: 'ok'
      };
    }
  }

  private  onPlayerReady(data: any): any {
    const { userId, sessionId } = data;
    const session: MatchSession | undefined = this.sessionManager.getSession(sessionId);
    if (session !== undefined) {
      session.setPlayerReady(userId)
      this.notifyService.notifyMatchState(session);
      this.notifyService.notifyPlayersState(session.players);
    }
  }

  private onClientReady(data: any): any {
    const { sessionId, userId } = data;
    const session: MatchSession | undefined = this.sessionManager.getSession(sessionId);
    session?.setClientReady(userId);
    return {
      status: 'ok'
    }
  }

  public updateSocketId(sessionId: string, userId: string, socketId: string): any {
    return this.sessionManager.updateSocketId(sessionId, userId, socketId);
  }

  
}